"""
Main data structures for the Runbot.
"""
import copy
import optparse
import os
import re
import simplejson
import subprocess
import sys
import threading
import time
import traceback

import mako.template

import openerprunbot
from openerprunbot.misc import *

# Hard-coded branches we always want, used to complement monitored branches
# (i.e. to create branch groups).
MAIN_BRANCHES = {
    'server_6.0': '~openerp/openobject-server/6.0',
    'server_trunk': '~openerp/openobject-server/trunk',
    'addons_6.0': '~openerp/openobject-addons/6.0',
    'addons_trunk': '~openerp/openobject-addons/trunk',
    'client_web_6.0': '~openerp/openobject-client-web/6.0',
    'web_trunk': '~openerp/openerp-web/trunk',
}

# Number of build 'slots' per branch group.
POINTS = 5

# This constant matches community_dashboard.py.
NEW_MERGE_STATUS = ['Needs review', 'Code failed to merge', 'Approved']

#----------------------------------------------------------
# OpenERP RunBot misc
#----------------------------------------------------------

def underscore(s):
    return s.replace("~","").replace(":","_").replace("/","_").replace(".","_")

def get_committer_info(repo_path):
    committer_name = None
    committer_xgram = None
    committer_email = None

    output = run_output(["bzr", "log", "--long", "-r-1"], cwd=repo_path)

    committer_re = re.compile('committer: *(.+)<(.+)@(.+)>')
    for i in output.split('\n'):
        m = committer_re.match(i)
        if m:
            committer_name = m.group(1).strip()
            committer_xgram = m.group(2)
            committer_email = m.group(2) + '@' + m.group(3)
            break

    return committer_name, committer_xgram, committer_email

#----------------------------------------------------------
# OpenERP RunBot Branch
#----------------------------------------------------------

class RunBotBranch(object):
    def __init__(self, runbot, branch, group=None,
        overriden_repo_path=None, overriden_project_name=None):
        """
        Represent a single Launchpad branch, e.g.
        `~openerp-dev/openobject-server/trunk-foo-bar`.

        A branch has an overriden_repo_path when it is a trunk or 6.0 branch
        used to complement another 'real' branch.
        """
        self.runbot = runbot
        self.group = group

        # e.g. trunk
        self.name = branch.name
        self.name_underscore = underscore(self.name)
        # e.g. ~openerp/openobject-server/trunk
        self.unique_name = branch.unique_name
        self.unique_name_underscore = underscore(self.unique_name)
        # e.g. server
        if overriden_project_name:
            self.project_name = overriden_project_name
        else:
            self.project_name = re.search("/(openobject|openerp)-(addons|server|client-web|web)/",self.unique_name).group(2)

        self.lp_date_last_modified = branch.date_last_modified
        self.lp_revision_count = branch.revision_count

        self.local_date_last_modified = 0
        self.local_revision_count = 0
        self.merge_count = 0

        # Repository path <Root>/repo/<branchuniquename>
        self.repo_path=os.path.join(self.runbot.wd,'repo',self.unique_name_underscore)
        self.overriden_repo_path = overriden_repo_path
        if overriden_repo_path:
            self.repo_path=overriden_repo_path

        self.committer_name = None
        self.committer_xgram = None
        self.committer_email = None

    def age(self):
        return time.time()-int(self.local_date_last_modified.strftime("%s"))

    def update_launchpad(self,branch):
        self.lp_date_last_modified=branch.date_last_modified
        self.lp_revision_count=branch.revision_count
        if self.local_revision_count != self.lp_revision_count:
            self.local_date_last_modified = self.lp_date_last_modified
            self.local_revision_count = self.lp_revision_count
            if not self.overriden_repo_path or self.group.is_configured():
                name = self.project_name
                self.group.need_run_reason.append(name)

#----------------------------------------------------------
# OpenERP RunBot Grouped Branch
#----------------------------------------------------------

class RunBotGroupedBranchBase(object):
    def __init__(self, runbot, sticky):
        self.runbot = runbot
        self.sticky = sticky
        self.points = [None for x in xrange(POINTS)]
        self.need_run_reason = []
        self.version = None
        self.wrong_matching = False

        # A normal manual_build is maxint, a manual_build is used only when
        # manually requesting a build with a 'build' command pushed in the
        # queue. The value will be reset to maxint after the build is done.
        self.manual_build = sys.maxint

        # List of modules to install instead of `all`.
        self.modules = []

    def add_point(self, j):
        assert not j.completed
        p = Point(self, j)
        self.points = self.points[1:] + [p] # TODO delete db and on-disk data

    def complete_point(self, j):
        assert j.completed
        for p in self.points + [None]:
            if p and p.job_id == j.job_id:
                break
        if p and p.state != 'running':
            p.update(j)
            p.save_json()

    def all_points_completed(self):
        for p in self.points:
            if p is not None and p.state != 'running':
                return False
        return True

    def update_state(self, state):
        """Copy some values taken from the state.runbot file."""
        if self.team_name == 'openerp-dev' and self.name in ('6.0', 'trunk'):
            # Don't change stickiness for the hard-coded branches.
            return
        previous = self.sticky
        self.sticky = state.get(self.team_name, {}).get(self.name, {}).get('sticky', False)
        if self.sticky != previous:
            log("stickiness changed", team=self.team_name, branch=self.name, now=self.sticky)

    def is_configured(self):
        return isinstance(self, ConfiguredGroup)

class ConfiguredGroup(RunBotGroupedBranchBase):
    def __init__(self, runbot, team_name, name, version, sticky,
            server_branch=None, client_web_branch=None,
            web_branch=None, addons_branches=None, modules=None):
        super(ConfiguredGroup, self).__init__(runbot, sticky)
        self.team_name = team_name
        self.name = name # Unique manually-chosen name, not necessarily the group name.
        self.name_underscore = underscore(self.name)
        self.version = version # '6.0' or 'trunk'

        self.server_branch = server_branch
        self.client_web_branch = client_web_branch
        self.web_branch = web_branch
        self.addons_branches = addons_branches

        self.server = None
        self.client_web = None
        self.web = None
        self.addons = []

        self.json_path=os.path.join(runbot.wd,'static',"%s-%s.json"%(team_name, self.name.replace('_','-').replace('.','-')))

        self.modules = modules.split(',') if modules else []

    def is_ok(self):
        """
        Test whether the group is useable, i.e. if self.populate_branches()
        didn't early return.
        """
        if self.server and (self.web or self.client_web) and self.addons:
            return True
        else:
            return False

    def populate_branches(self, launchpad):
        log("runbot-populate-configured-branches")
        repo = os.path.join(self.runbot.wd, 'repo')
        filters = {'status': NEW_MERGE_STATUS}

        path = os.path.join(repo, '__configured_' + self.name + '_server')
        b = launchpad.get_branch(unique_name=self.server_branch)
        if not b:
            log("WARNING:no such unique name", name=self.server_branch)
            return
        self.server = RunBotBranch(self.runbot, b, group=self, overriden_repo_path=path, overriden_project_name='server')
        self.server.update_launchpad(b)
        self.server.merge_count = len(list(b.getMergeProposals(**filters)))

        if self.web_branch:
            path = os.path.join(repo, '__configured_' + self.name + '_web')
            b = launchpad.get_branch(unique_name=self.web_branch)
            if not b:
                log("WARNING:no such unique name", name=self.web_branch)
                return
            self.web = RunBotBranch(self.runbot, b, group=self, overriden_repo_path=path, overriden_project_name='web')
            self.web.update_launchpad(b)
            self.web.merge_count = len(list(b.getMergeProposals(**filters)))
        else:
            self.web = None

        if self.client_web_branch:
            path = os.path.join(repo, '__configured_' + self.name + '_web')
            b = launchpad.get_branch(unique_name=self.client_web_branch)
            if not b:
                log("WARNING:no such unique name", name=self.client_web_branch)
                return
            self.client_web = RunBotBranch(self.runbot, b, group=self, overriden_repo_path=path, overriden_project_name='client-web')
            self.client_web.update_launchpad(b)
            self.client_web.merge_count = len(list(b.getMergeProposals(**filters)))
        else:
            self.client_web = None

        self.addons = []
        for x in xrange(len(self.addons_branches)):
            path = os.path.join(repo, '__configured_' + self.name + '_addons_' + str(x+1) + 'of' + str(len(self.addons)))
            b = launchpad.get_branch(unique_name=self.addons_branches[x])
            if not b:
                log("WARNING:no such unique name", name=self.addons_branches[x])
                self.addons = []
                return
            bb = RunBotBranch(self.runbot, b, group=self, overriden_repo_path=path, overriden_project_name='addons')
            bb.update_launchpad(b)
            bb.merge_count = len(list(b.getMergeProposals(**filters)))
            self.addons.append(bb)

    def repo_updates(self):
        r = []
        r.append(copy.copy(self.server))
        for x in self.addons:
            r.append(copy.copy(x))
        if self.client_web:
            r.append(copy.copy(self.client_web))
        if self.web:
            r.append(copy.copy(self.web))
        return r

    def lp_date_last_modified(self):
        lp_date_last_modified = None
        for p in [self.server, self.web] + self.addons:
            if p and (not lp_date_last_modified or p.lp_date_last_modified > lp_date_last_modified):
                lp_date_last_modified = p.lp_date_last_modified
        assert lp_date_last_modified # at least one project in the group
        return lp_date_last_modified

class RunBotGroupedBranch(RunBotGroupedBranchBase):
    """
    A single 'branch group' represents a single branch name for all our
    projects. E.g. trunk is available in server, addons, web, and client-web.

    When the corresponding project doesn't have that branch, trunk is used.
    For instance, a branch group can represent
        `~openerp-dev/openobject-server/trunk-foo-bar`
        `~openerp-dev/openobject-addons/trunk-foo-bar`
    and
        `~openerp/openobject-web-client/6.0`
        `~openerp/openerp-web/trunk`
    will be used to complete the group, building the four branches together.

    Alternatively a 'branch group' can be configured. In that case, all
    branches are manually specified; no attempt is done to complete the group.
    """

    def __init__(self, runbot, b, team_name, sticky):
        """A grouped branch can be created from any branch."""
        super(RunBotGroupedBranch, self).__init__(runbot, sticky)
        self.team_name = team_name
        self.name = b.name
        self.name_underscore = underscore(self.name)
        self.server = None
        self.addons = [] # Only one branch in a non-configured group.
        self.web = None
        self.client_web = None
        self.json_path=os.path.join(runbot.wd,'static',"%s-%s.json"%(team_name, self.name.replace('_','-').replace('.','-')))

        self.add_branch(b)

    def add_branch(self, b):
        assert b.name == self.name
        b.group = self
        if b.project_name == 'addons':
            self.addons = [b]
        if b.project_name == 'server':
            self.server = b
        if b.project_name == 'client-web':
            self.client_web = b
        if b.project_name == 'web':
            self.web = b

    # Note: any access to launchpad is done in the same thread.
    def add_main_branches(self):
        """
        Add trunk or 6.0 branches when the complementary branches are missing.
        Also do it when the addons branch is not complete, i.e. when missing
        one of the official addons.
        """
        repo = os.path.join(self.runbot.wd,'repo')

        if self.name.startswith('6.0'):
            ending = "_6.0"
            self.version = '6.0'
        elif self.name.startswith('trunk'):
            ending = "_trunk"
            self.version = 'trunk'
        else:
            self.wrong_matching = True
            ending = "_trunk"
            self.version = 'trunk' # TODO this is not accurate

        paths = {
            'addons': os.path.join(repo, 'openerp_openobject-addons_' + self.name_underscore),
            'server': os.path.join(repo, 'openerp_openobject-server_' + self.name_underscore),
            'client_web': os.path.join(repo, 'openerp_openobject-client-web_' + self.name_underscore),
            'web': os.path.join(repo, 'openerp_openerp-web_' + self.name_underscore),
        }

        for p in ('server', 'client_web', 'web'):
            if not getattr(self, p):
                b = self.runbot.main_branches[p+ending]
                setattr(self, p, RunBotBranch(self.runbot,b,group=self,overriden_repo_path=paths[p]))
                getattr(self, p).update_launchpad(b)
            elif getattr(self, p).overriden_repo_path:
                b = self.runbot.main_branches[p+ending]
                getattr(self, p).update_launchpad(b)
        if not self.addons:
            b = self.runbot.main_branches['addons'+ending]
            self.addons = [RunBotBranch(self.runbot,b,group=self,overriden_repo_path=paths['addons'])]
            self.addons[0].update_launchpad(b)
        elif self.addons[0].overriden_repo_path:
            b = self.runbot.main_branches['addons'+ending]
            self.addons[0].update_launchpad(b)

    def repo_updates(self):
        r = []
        r.append(copy.copy(self.server))
        r.append(copy.copy(self.addons[0]))
        r.append(copy.copy(self.client_web))
        r.append(copy.copy(self.web))
        return r

    def lp_date_last_modified(self):
        lp_date_last_modified = None
        for p in ('server', 'client_web', 'web'):
            if getattr(self, p) and not getattr(self, p).overriden_repo_path \
                and (not lp_date_last_modified or getattr(self, p).lp_date_last_modified > lp_date_last_modified):
                lp_date_last_modified = getattr(self, p).lp_date_last_modified
        if self.addons and not self.addons[0].overriden_repo_path \
            and (not lp_date_last_modified or self.addons[0].lp_date_last_modified > lp_date_last_modified):
            lp_date_last_modified = self.addons[0].lp_date_last_modified
        assert lp_date_last_modified # at least one project in the group
        return lp_date_last_modified

#----------------------------------------------------------
# OpenERP RunBot Worker
#----------------------------------------------------------

class Job(object):
    """
    A Job encapsulates all the necessary data to build a branch group for a
    given slot. The build is done in its own worker thread.
    """

    def __init__(self, g, port, test, job_id, debug):
        self.job_id = job_id
        self.completed = False
        self.team_name = g.team_name
        self.name = g.name
        self.name_underscore = g.name_underscore
        self.version = g.version
        self.debug = debug

        self.repo_updates = g.repo_updates()
        self.port = port
        self.test = test
        self.running_server_pid = 0
        self.client_web_pid = 0
        self.running_t0=0

        repo = os.path.join(g.runbot.wd,'repo')
        self.server_src = g.server.repo_path
        self.client_web_src = g.client_web.repo_path if g.client_web else None
        self.web_src = g.web.repo_path if g.web else None

        # if addons is not the full addons branch use trunk
        self.addons_src = [a.repo_path for a in g.addons]

        # Running path <Root>/static/<domain>
        self.subdomain = "%s-%s-%s"%(self.team_name, self.name.replace('_','-').replace('.','-'),self.job_id)
        self.running_path = os.path.join(g.runbot.wd,'static',self.subdomain)
        self.json_path = g.json_path
        self.log_path = os.path.join(self.running_path,'logs')
        self.flags_path = os.path.join(self.running_path,'flags')

        # Server
        self.server_path=os.path.join(self.running_path,"server")
        self.server_bin_path=os.path.join(self.server_path,"openerp-server")
        self.server_log_path=os.path.join(self.log_path,'server.txt')
        self.server_log_base_path=os.path.join(self.log_path,'test-base.txt')
        self.server_log_all_path=os.path.join(self.log_path,'test-all.txt')

        # coverage
        self.coverage_file_path=os.path.join(self.log_path,'coverage.pickle')
        self.coverage_base_path=os.path.join(self.log_path,'coverage-base')
        self.coverage_all_path=os.path.join(self.log_path,'coverage-all')

        # Web60
        self.client_web_pid=None
        self.client_web_path=os.path.join(self.running_path,"client-web")
        self.client_web_bin_path=os.path.join(self.client_web_path,"openerp-web.py")
        self.client_web_doc_path=os.path.join(self.client_web_path,"doc")
        self.client_web_log_path=os.path.join(self.log_path,'client-web.txt')

        # test
        self.test_base_result=None
        self.test_base_path=os.path.join(self.log_path,'test-base.txt')
        self.test_all_result=None
        self.test_all_path=os.path.join(self.log_path,'test-all.txt')

        self.server_net_port=g.runbot.server_net_port
        self.server_xml_port=g.runbot.server_xml_port
        self.client_web_port=g.runbot.client_web_port

        self.db = self.name_underscore + '_' + str(self.job_id)
        self.db_all = "%s_all" % self.db

        for p in ('addons', 'server', 'client_web', 'web'):
            setattr(self, p + '_committer_name', None)
            setattr(self, p + '_committer_xgram', None)
            setattr(self, p + '_committer_email', None)
        self.start_time = 0
        self.completed_time = 0

        self.modules = g.modules

    def spawn(self):
        log("runbot-spawn-worker-" + str(self.job_id), group=self.name)
        t = threading.Thread(target=self.work, name=('runbot-group-worker-' + str(self.job_id)))
        t.daemon = True
        t.start()

    def work(self):
        try:
            self.pull_branches()
            self.start_time = time.time()
            self.start()
            self.completed_time = time.time()
            self.completed = True
            log("runbot-end-worker", job=self.name)
        except Exception, e:
            self.completed = True
            log("runbot-end-worker [with exception]", job=self.name)
            print traceback.format_exc()

    def pull_branches(self):
        for b in self.repo_updates:
            log("branch-update",branch=b.unique_name)
            if os.path.exists(b.repo_path):
                run(["bzr", "pull", "-d", b.repo_path, "--overwrite"])
            else:
                run(["bzr", "branch", "lp:%s"%b.unique_name, b.repo_path])
            run(["bzr", "update", "-r", str(b.local_revision_count), b.repo_path])

            committer_name, committer_xgram, committer_email = \
                get_committer_info(b.repo_path)
            b.committer_name = committer_name
            b.committer_xgram = committer_xgram
            b.committer_email = committer_email

    def start_rsync(self):
        log("job-start-rsync",branch=self.name)
        for i in [self.running_path,self.log_path,self.flags_path,self.client_web_doc_path]:
            if not os.path.exists(i):
                os.makedirs(i)
        # copy server
        run(["rsync","-a","--exclude",".bzr","--delete", "%s/"%self.server_src, self.server_path])

        # copy addons (6.0 uses bin/addons 6.1 openerp/addons)
        addons_dest = "openerp/addons"
        if not os.path.exists(os.path.join(self.server_path,addons_dest)):
            addons_dest = "bin/addons"
        addons_path = os.path.join(self.server_path,addons_dest)
        for a in self.addons_src:
            run(["rsync","-a","--exclude",".bzr", "%s/"%a, addons_path])
        if self.web_src:
            run(["rsync","-a","--exclude",".bzr", "%s/addons/"%self.web_src, addons_path])

        # copy web-client
        if self.client_web_src:
            run(["rsync","-a","--exclude",".bzr","--delete", "%s/"%self.client_web_src, self.client_web_path])

    def start_createdb(self):
        run(["psql","template1","-c","select pg_terminate_backend(procpid) from pg_stat_activity where datname in ('%s','%s')"%(self.db,self.db_all)])
        time.sleep(3)
        run(["dropdb",self.db])
        run(["dropdb",self.db_all])
        run(["createdb",self.db])
        run(["createdb",self.db_all])

    def run_log(self, cmd, logfile, env=None):
        env = dict(os.environ, **env) if env else None
        log("run", *cmd, logfile=logfile)
        out=open(logfile,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True, env=env)
        self.running_server_pid=p.pid
        p.communicate()
        return p

    def resolve_server_bin_path(self):
        # This can be done only if the files are present otherwise the if
        # will always fail. Alternatively, server_bin_path could be a property.
        if not os.path.exists(self.server_bin_path): # for 6.0 branches
            self.server_bin_path=os.path.join(self.server_path,"bin","openerp-server.py")

    def start_test_base(self):
        log("job-start-server-base")
        cmd = [self.server_bin_path,"-d",self.db,"-i","base","--stop-after-init","--no-xmlrpc","--no-xmlrpcs","--no-netrpc","--log-level=test"]
        cmd = ["coverage","run","--branch"] + cmd
        self.run_log(cmd, logfile=self.test_base_path,env={'COVERAGE_FILE': self.coverage_file_path})
        run(["coverage","html","-d",self.coverage_base_path,"--ignore-errors","--include=*.py"],env={'COVERAGE_FILE': self.coverage_file_path})
        rc = run(["grep","Traceback",self.test_base_path])
        self.test_base_result = bool(rc)

    def start_test_all(self):
        log("job-start-server-all")
        mods = []
        if not os.path.exists(os.path.join(self.server_path,"openerp/addons")):
            mods = os.listdir(os.path.join(self.server_path,"bin/addons"))
        elif os.path.exists(os.path.join(self.server_path,"openerp/addons")):
            mods = os.listdir(os.path.join(self.server_path,"openerp/addons"))
        bl = [
            'document_ftp', 'l10n_lu',
            '.bzrignore', '__init__.pyc', '__init__.py', 'base_quality_interrogation.py',
        ]
        mods = [m for m in mods if m not in bl]
        if self.modules:
            # TODO feedback when a requested modules does'nt exist.
            mods = [m for m in mods if m in self.modules]
        mods = ",".join(mods)
        cmd = [self.server_bin_path,"-d",self.db_all,"-i",mods,"--stop-after-init","--no-xmlrpc","--no-xmlrpcs","--no-netrpc","--log-level=test"]
        cmd = ["coverage","run","--branch"] + cmd
        self.run_log(cmd, logfile=self.test_all_path)
        run(["coverage","html","-d",self.coverage_all_path,"--ignore-errors","--include=*.py"])
        self.test_all_result = bool(run(["grep","Traceback",self.test_all_path]))

    def start_server(self):
        port = self.port
        log("job-start-server",branch=self.name,port=port)
        cmd=[self.server_bin_path,"--no-xmlrpcs","--netrpc-port=%d"%(self.server_net_port+port),"--xmlrpc-port=%d"%(self.server_xml_port+port)]
        if os.path.exists(os.path.join(self.server_path, 'openerp', 'wsgi.py')) \
            or os.path.exists(os.path.join(self.server_path, 'openerp', 'wsgi', 'core.py')):
            cmd+=["--db-filter=%d(_.*)?$","--load=web"]
        log("run",*cmd,log=self.server_log_path)
        out=open(self.server_log_path,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True)
        self.running_server_pid=p.pid

    def start_client_web(self):
        if not self.client_web_src:
            return
        port = self.port
        log("job-start-client-web",branch=self.name,port=port)
        config="""
        [global]
        server.environment = "development"
        server.socket_host = "0.0.0.0"
        server.socket_port = %d
        server.thread_pool = 10
        tools.sessions.on = True
        log.access_level = "INFO"
        log.error_level = "INFO"
        tools.csrf.on = False
        tools.log_tracebacks.on = False
        tools.cgitb.on = True
        openerp.server.host = 'localhost'
        openerp.server.port = '%d'
        openerp.server.protocol = 'socket'
        openerp.server.timeout = 450
        [openerp-web]
        dblist.filter = 'BOTH'
        dbbutton.visible = True
        company.url = ''
        openerp.server.host = 'localhost'
        openerp.server.port = '%d'
        openerp.server.protocol = 'socket'
        openerp.server.timeout = 450
        """%(self.client_web_port+port,self.server_net_port+port,self.server_net_port+port)
        config=config.replace("\n        ","\n")
        cfgs = [os.path.join(self.client_web_path,"doc","openerp-web.cfg"), os.path.join(self.client_web_path,"openerp-web.cfg")]
        for i in cfgs:
            f=open(i,"w")
            f.write(config)
            f.close()

        cmd=[self.client_web_bin_path]
        log("run",*cmd,log=self.client_web_log_path)
        out=open(self.client_web_log_path,"w")
        p=subprocess.Popen(cmd, stdout=out, stderr=out, close_fds=True)
        self.client_web_pid=p.pid

    def start(self):
        log("job-start",branch=self.name,port=self.port)
        self.start_rsync()
        self.resolve_server_bin_path()
        self.start_createdb()
        try:
            if self.test:
                self.start_test_base()
                if not self.debug:
                    self.start_test_all()
                else:
                    self.test_all_result = True
            self.start_server()
            self.start_client_web()
        except OSError,e:
            log("branch-start-error",exception=e)
        except IOError,e:
            log("branch-start-error",exception=e)
        self.running_t0=time.time()
        log("branch-started",branch=self.name,port=self.port)

    def stop(self):
        log("job-stop",branch=self.name)
        if self.running_server_pid:
            kill(self.running_server_pid)
        if self.client_web_pid:
            kill(self.client_web_pid)

#----------------------------------------------------------
# OpenERP RunBot Build Slot
#----------------------------------------------------------

class Point(object):
    """A point is a build slot and associated to a worker thread."""
    def __init__(self, g, j):
        """Create a Point from a given group and job."""
        self.version = g.version
        self.port = j.port
        self.job_id = j.job_id
        self.db = j.db
        self.running_path = j.running_path
        self.json_path = j.json_path
        self.subdomain = j.subdomain
        self.repo_updates = j.repo_updates # committer date not yet available, available after j.work() is called.
        self.server_rev = g.server.local_revision_count
        self.addons_rev = [a.local_revision_count for a in g.addons]
        self.client_web_rev = g.client_web.local_revision_count if g.client_web else 0
        self.web_rev = g.web.local_revision_count if g.web else 0
        self.need_run_reason = g.need_run_reason
        for p in ('addons', 'server', 'client_web', 'web'):
            setattr(self, p + '_committer_name', getattr(j, p + '_committer_name'))
            setattr(self, p + '_committer_xgram', getattr(j, p + '_committer_xgram'))
            setattr(self, p + '_committer_email', getattr(j, p + '_committer_email'))
        self.update(j)
        self.manual_build = g.manual_build

    def update(self, j):
        """
        Update the Point from a job. The job should be the same than the one
        used in `__init__()`.
        """
        self.state = 'running' if j.completed else 'testing'
        self.running_t0 = j.running_t0
        self.test_base_result = j.test_base_result
        self.test_all_result = j.test_all_result

    def save_json(self):
        """Append the point data to a JSON file for posterity."""
        # A job must be saved only once.
        if hasattr(self, 'json_saved'):
            log("=== save_json() called more than once. ===", job_id=self.job_id)
        self.json_saved = True

        path = self.json_path
        state = {}

        if os.path.exists(path):
            with open(path, 'r') as h:
                state = simplejson.loads(h.read())

        value = {}
        committers = []
        for p in ('addons', 'server', 'client_web', 'web'):
            committers += [
                p + '_committer_name',
                p + '_committer_xgram',
                # p + '_committer_email',
            ]
        for a in [
            'job_id', 'db', 'running_path', 'subdomain', 'server_rev',
            'addons_rev', 'client_web_rev', 'web_rev', 'need_run_reason',
            'test_base_result', 'test_all_result',
            ] + committers:
            value[a] = getattr(self, a)
        state.setdefault('jobs', [])
        state['jobs'].append(value)

        with open(path, 'w') as h:
            data = simplejson.dumps(state, sort_keys=True, indent=4)
            h.write(data)

#----------------------------------------------------------
# OpenERP RunBot Engine
#----------------------------------------------------------

class RunBot(object):
    """Used as a singleton, to manage almost all the Runbot state and logic."""
    def __init__(self, wd, poll, server_net_port, server_xml_port,
        client_web_port,  number, nginx_port, domain, test, workers,
        current_job_id, debug, lp):
        self.wd=wd
        self.server_net_port=int(server_net_port)
        self.server_xml_port=int(server_xml_port)
        self.client_web_port=int(client_web_port)
        self.number=int(number)
        self.nginx_port=int(nginx_port)
        self.domain=domain
        self.test=int(test)
        self.workers = int(workers)
        self.branches={}
        self.groups={}
        self.main_branches = {}
        self.allocated_port = 0
        self.jobs = {}
        self.current_job_id = current_job_id
        self.debug = debug
        self.manual_build_count = 0
        self.launchpad = lp

    def registered_teams(self):
        return openerprunbot.state.get('registered-teams', []) + ['openerp-dev']

    def nginx_reload(self):
        nginx_pid_path = os.path.join(self.wd,'nginx','nginx.pid')
        if os.path.isfile(nginx_pid_path):
            pid=int(open(nginx_pid_path).read())
            os.kill(pid,1)
        else:
            run(["nginx","-p",self.wd,"-c",os.path.join(self.wd,"nginx/nginx.conf")])

    def nginx_config(self):
        return openerprunbot.templates.render_template('nginx.conf.mako',r=self)

    def registration_page(self):
        return openerprunbot.templates.render_template('registration.html.mako',r=self)

    def nginx_index_time(self,t):
        for m,u in [(86400,'d'),(3600,'h'),(60,'m')]:
            if t>=m:
                return str(int(t/m))+u
        return str(int(t))+"s"

    def nginx_index_sticky(self):
        #branches = [(b.subdomain,b) for b in self.running if b.sticky]
        #branches.sort()
        #return [b for (x,b) in branches]
        return []

    def nginx_groups_sticky(self,team_name):
        gs = [(g.name,g) for g in self.groups.values() if any(g.points) and g.team_name==team_name and g.sticky]
        gs.sort()
        return [g for (x,g) in gs]

    def nginx_groups_others(self,team_name):
        gs = [(g.sticky, g.lp_date_last_modified(), g) for g in self.groups.values() if any(g.points) and g.team_name==team_name and not g.sticky]
        gs.sort(reverse=1)
        gs = [g for (x,y,g) in gs]
        return gs

    def nginx_groups_registered(self,team_name):
        gs = [(g.manual_build, g.name,g) for g in self.groups.values() if not any(g.points) and g.team_name==team_name]
        gs.sort()
        return [g for (x,y,g) in gs]

    def nginx_index_others(self):
        #branches = [(b.local_date_last_modified,b) for b in self.running if not b.sticky]
        #branches.sort(reverse=1)
        #return [b for (x,b) in branches]
        return []

    def nginx_index(self, team_name):
        return openerprunbot.templates.render_template('branches.html.mako',
            r=self, t=time.time(), re=re, team_name=team_name, sys=sys)

    def nginx_teams(self):
        return openerprunbot.templates.render_template('index.html.mako',
            r=self,t=time.time(),re=re)

    def nginx_update(self):
        try:
            f = None
            f = open(os.path.join(self.wd,'static','index.html'),"w")
            content = self.nginx_teams()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating index.html:")
            print traceback.format_exc()
            if f: f.close()
        for team_name in self.teams_with_branches():
            try:
                f = None
                fn = team_name + '.html'
                f = open(os.path.join(self.wd,'static',fn),"w")
                content = self.nginx_index(team_name)
                f.write(content)
            except Exception, e:
                log("WARNING: exception when templating %s:" % fn)
                print traceback.format_exc()
                if f: f.close()
                break

        try:
            f = open(os.path.join(self.wd,'static','register.html'),"w")
            content = self.registration_page()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating register.html:")
            print traceback.format_exc()
            if f: f.close()

        try:
            f = open(os.path.join(self.wd,'nginx','nginx.conf'),"w")
            content = self.nginx_config()
            f.write(content)
        except Exception, e:
            log("WARNING: exception when templating nginx.conf:")
            print traceback.format_exc()
            if f: f.close()
        self.nginx_reload()

    def teams_with_branches(self):
        ts = [g.team_name for g in self.groups.values()]
        ts = list(set(ts))
        ts.sort()
        return ts

    def allocate_port(self):
        if self.allocated_port > self.number * 2:
            self.allocated_port = 0
        self.allocated_port += 1
        return self.allocated_port

    def process_add(self, b, sticky, team_name):
        log("runbot-process-add", b.unique_name)
        if not re.search("/(openobject|openerp)-(addons|server|client-web|web)/", b.unique_name):
            log("WARNING: can't add branch: project name is not standard.", unique_name=b.unique_name)
            return
        if b.unique_name not in self.branches:
            self.branches[b.unique_name] = RunBotBranch(self, b)
        bb = self.branches[b.unique_name]
        if (team_name, bb.name) not in self.groups:
            self.groups[(team_name, bb.name)] = RunBotGroupedBranch(self, bb, team_name, sticky)
        else:
            self.groups[(team_name, bb.name)].add_branch(bb)
        bb.update_launchpad(b)
        filters = {'status': NEW_MERGE_STATUS}
        bb.merge_count = len(list(b.getMergeProposals(**filters)))

    def register_configured_branches(self):
        for team, v in openerprunbot.state.get('configured-branches', {}).items():
            if not team: continue
            for name, c in v.items():
                if not name: continue
                g = ConfiguredGroup(self, team, name, c['version'], 0,
                    server_branch=c['server_branch'],
                    client_web_branch=c.get('client_web_branch'),
                    web_branch=c.get('web_branch'),
                    addons_branches=c['addons_branches'],
                    modules=c['modules'])

                # TODO or if it is already there but with different branches.
                if (g.team_name, g.name) not in self.groups:
                    g.populate_branches(self.launchpad)
                    if g.is_ok():
                        log("adding configured group", team=g.team_name, name=g.name)
                        self.groups[(g.team_name, g.name)] = g
                    else:
                        log("WARNING: misconfigured group", team=g.team_name, name=g.name)

    # Note: any access to launchpad is done in the same thread.
    def populate_branches(self):
        """Return all LP branches matching our teams and projects."""
        log("runbot-populate-branches")
        # Register main sticky branches
        for k, v in MAIN_BRANCHES.items():
            b = self.launchpad.get_branch(unique_name=v)
            self.process_add(b, 1, 'openerp-dev')
            self.main_branches[k] = b
        self.main_branches['client_web_trunk'] = self.main_branches['client_web_6.0']
        self.main_branches['web_6.0'] = self.main_branches['web_trunk']

        # Register other branches
        for v in openerprunbot.state.get('registered-branches', []):
            m = openerprunbot.branch_input_re.match(v)
            if m:
                b = self.launchpad.get_branch(unique_name=v)
                if not b:
                    # TODO better handling of launchpad connection problem
                    log("WARNING: can't get branch:", name=v)
                    continue
                self.process_add(b, 0, m.group(1))
            else:
                log("WARNING: malformed branch name:", v)

        # Register team branches
        for team_name in self.registered_teams():
            log("Register team branches:", team=team_name)
            if self.debug: break
            team_branches = self.launchpad.get_team_branches(team_name)
            for b in team_branches:
                if re.search("/(openobject|openerp)-(addons|server|client-web|web)/",b.unique_name):
                    self.process_add(b, 0, team_name)

        # Register configured branches
        self.register_configured_branches()

        for g in self.groups.values():
            if not g.is_configured():
                g.add_main_branches()

        return self.sorted_groups()

    def sorted_groups(self):
        gs = [(g.sticky, -g.manual_build, g.lp_date_last_modified(), g) for g in self.groups.values()]
        gs.sort(reverse=1)
        gs = [g for (x,y,z,g) in gs][:self.number]
        return gs

    def available_workers(self):
        return self.workers - len([t for t in threading.enumerate() if t.name.startswith('runbot-group-worker-')])

    def process_command_queue(self):
        while not openerprunbot.queue.empty():
            command, params = openerprunbot.queue.get()
            if command == 'build':
                team_name, group_name = params
                if (team_name, group_name) in self.groups and self.groups[(team_name, group_name)].manual_build == sys.maxint:
                    self.manual_build_count += 1
                    self.groups[(team_name, group_name)].manual_build = self.manual_build_count
                    self.groups[(team_name, group_name)].need_run_reason.append('build')
            else:
                log("WARNING: unknown command", command)

    def reset_build_numbers(self):
        gs = [(g.manual_build, g) for g in self.groups.values()]
        gs.sort()
        gs = [g for (x,g) in gs]
        self.manual_build_count = 0
        for g in gs:
            if g.manual_build == sys.maxint:
                break
            self.manual_build_count += 1
            g.manual_build = self.manual_build_count

    def complete_jobs(self):
        """Update all slots with the completed jobs."""
        for job in self.jobs.values():
            if job.completed:
                for g in self.groups.values():
                    if g.name == job.name:
                        g.complete_point(job)
                        break

    def reap_oldest_job(self):
        """Kill some job to limit the number of processes."""
        if len(self.jobs) > self.number:
            victim = None
            for job in self.jobs.itervalues():
                if not victim or job.job_id < victim.job_id:
                    victim = job
            if victim:
                victim.stop()
                del self.jobs[victim.job_id]

    def fetch_groups(self):
        """
        Fetch branch information from Launchpad.
        """
        log("runbot-process-grouped-branches")

        if self.available_workers():
            self.launchpad.connect()
            gs = self.populate_branches()
        else:
            gs = []

        log("runbot-run-group", groups=len(gs))
        return gs

    def run_groups(self, gs):
        for g in gs:
            if not self.available_workers():
                break
            if len(self.jobs) > self.number:
                break
            # Don't run multiple jobs at the same time for the same group.
            if g.need_run_reason and g.all_points_completed():
                port = self.allocate_port()
                self.current_job_id += 1
                job = Job(g, port, self.test, self.current_job_id, self.debug)
                g.add_point(job)
                g.need_run_reason = []
                g.manual_build = sys.maxint
                self.jobs[job.job_id] = job
                job.spawn()

    def loop(self):
        """
        Repeatedly fetch branch information from Launchpad, run jobs, update
        HTML pages, ...
        """
        while 1:
            try:
                gs = []
                try:
                    gs = self.fetch_groups()
                except Exception, e:
                    log("WARNING: exception:")
                    print traceback.format_exc()

                for i in xrange(12): # 2 minutes
                    try: # Make sure time.sleep() is always called.
                        self.run_groups(gs)
                        self.complete_jobs()
                        self.reap_oldest_job()
                        self.nginx_update()
                        self.process_command_queue()
                        self.reset_build_numbers()
                        for g in self.groups.values():
                            g.update_state(openerprunbot.state)
                    except Exception, e:
                        log("WARNING: exception:")
                        print traceback.format_exc()
                    time.sleep(10)
            except KeyboardInterrupt,e:
                log("SIGINT received, exiting...")
                openerprunbot.server.stop_server()
                for i in self.jobs.values():
                    i.stop()
                break

